#!/usr/bin/env python

import atexit
import os
import sys
import subprocess
import ssl
import re
import getpass
from datetime import datetime
from subprocess import Popen, PIPE

sys.path.append("/usr/lib/vmware-marvin/marvind/webapps/ROOT/WEB-INF/classes/scripts/lib/python2.7/site-packages")

from pyVim.connect import Connect, Disconnect
from pyVmomi import vim, vmodl
from distutils.version import StrictVersion, LooseVersion

# VxRail Version Milestone For VC Migration
VXRAILNO_V0 = '4.0.400'
VXRAILNO_V3 = '4.7.000'
VXRAILNO_V4 = '4.7.100'
VXRAILNO_V5 = '4.7.300'
VIM_VERSION = "vim.version.version10"

# VxRail Plugin Properties
vxm_thumbprint_key_name = "VxRail-SSL-THUMBPRINT"
vxm_ip_key_name = "VxRail-IP"
vxm_version_key_name = "VxRail-VERSION"

RUNTIME_PROPERTIES_FILE = '/var/lib/vmware-marvin/runtime.properties'
PROPERTY_EXTENSION_KEY = 'data.extension.key'
PROPERTY_EXTENSION_VERSION = 'data.extension.version'

OKGREEN = '\033[92m'
ERROR = '\033[91m'
ENDC = '\033[0m'
SELECTION = '\033[93m'
DEBUG = '\033[94m'


def print_info(msg):
    """print executing information """
    print OKGREEN + msg + ENDC


def print_debug(msg):
    """print debugging information """
    print DEBUG + '[Debug]' + msg + ENDC


def print_selection(msg):
    """print selection option """
    print SELECTION + msg + ENDC


def print_error(msg):
    """print error information """
    print ERROR + msg + ENDC


def checkroot():
    """check root user """
    f_handle = os.popen('whoami')
    currentuser = (f_handle.readlines()[0].strip('\n'))
    if currentuser != "root":
        print "Current user is %s, please su to root." % currentuser
        exit(-1)


def execute_mystic_psql(cmd):
    """execute operations on mysticmanager database in VxRail Manager"""
    print_debug(cmd)
    f_handle = os.popen("psql -t -U postgres mysticmanager -c \"%s\"" % (cmd))
    result = f_handle.readlines()[0].strip(' \n ')
    print_debug(result)
    return result


def print_selection_menu(lst, selection_type):
    """print selection menu """
    while True:
        print_selection("Found more than 1 %s, please select the one you want to use" % (selection_type))
        for counter, value in enumerate(lst, 1):
            print "%d) %s" % (counter, value)
        print "0) Exit                   "
        x_selection = int(raw_input("Select an option[0-%d]:" % (len(lst))))

        if x_selection in range(0, len(lst) + 1):
            break
        else:
            os.system('clear')
    if x_selection == 0:
        exit(1)
    return x_selection - 1


def iterate_cluster_folder(items, nlist, elist, folder):
    """iterate cluster folder """
    for item in items:
        if isinstance(item, vim.ClusterComputeResource):
            if folder == "":
                nlist.append(item.name)
            else:
                nlist.append(folder + '\\' + item.name)
            elist.append(item)
        elif isinstance(item, vim.Folder):
            if folder == "":
                folder_inner = item.name
            else:
                folder_inner = folder + '\\' + item.name
            iterate_cluster_folder(item.childEntity, nlist, elist, folder_inner)


def collect_cluster(data_center, info):
    """get cluster """
    clusters = data_center.hostFolder.childEntity
    clu_list = []
    ent_clus = []
    iterate_cluster_folder(clusters, clu_list, ent_clus, "")

    if len(clu_list) < 1:
        print_error("Could not find validated cluster")
        exit(1)
    elif len(clu_list) == 1:
        index = 0
    else:
        index = print_selection_menu(clu_list, "cluster")

    ent_cluster = ent_clus[index]
    info['cluster_id'] = ent_cluster._moId
    info['cluster_name'] = ent_cluster.name
    print_info("Found cluster: [moId] %s [Name] %s" % (ent_cluster._moId,
                                                       ent_cluster.name))
    return ent_cluster


def collect_datacenter(content, info):
    """get datacenter """
    dcs = []
    dclist = []
    dc_view = content.viewManager.CreateContainerView(content.rootFolder,
                                                      [vim.Datacenter], True)
    for child in dc_view.view:
        dclist.append(child.name)
        dcs.append(child)
    dc_view.Destroy()
    if len(dcs) < 1:
        print_error("Could not find validated datacenter")
        exit(1)
    elif len(dcs) == 1:
        infoconfirm = raw_input('Please confirm to use datacenter %s on %s ?  y - continue, n - exit [default:y]:'
                                % (dclist[0], info['VC_IP']))
        if infoconfirm == 'y' or infoconfirm == '':
            index = 0
        else:
            exit(1)
    else:
        index = print_selection_menu(dclist, "datacenter")
    ent_datacenter = dcs[index]
    print_info("Found datacenter: [moId] %s [Name] %s" %
               (ent_datacenter._moId, ent_datacenter.name))
    info['dc_id'] = ent_datacenter._moId
    info['dc_name'] = ent_datacenter.name
    return ent_datacenter


def vxrailversion():
    """Check VxRail Version"""
    build_num = os.popen("rpm -qa | grep '^vxrail-mystic-' | cut -d- -f3")
    version_num = build_num.readlines()[0].strip(' \n ')

    if StrictVersion(version_num) >= StrictVersion(VXRAILNO_V4):
        return version_num
    else:
        print_error("WARNING: Current VxRail Version is %s, NOT Supported" % version_num)
        exit(-1)


def vxrail_plugin_check(vxrail_version, extension_manager):
    # Checking the followings:
    # 1. VxRail version about whether it's necessary to install/register VxRail plugin to Target vCenter
    # Server. it's only necessary when starting from VxRail No. 4.7.100
    # 2. VxRail plugin is installed or not in target vCenter Server.

    vxrail_extension = []

    for item in extension_manager.extensionList:
        if re.split('[_]', item.key)[0] == 'com.vmware.vxrail':
            vxrail_extension.append(item.key)

    if StrictVersion(vxrail_version) >= StrictVersion(VXRAILNO_V3):
        if len(vxrail_extension) >= 1:
            print_debug("Plugin VxRail Extensions %s." % vxrail_extension)
            plugin_check_result = 'Installed'
            return plugin_check_result, vxrail_extension

        elif len(vxrail_extension) == 0:
            print_info("VxRail Plugin check pass: no VxRail plugin installed")
            plugin_check_result = 'Pass'
            return plugin_check_result, vxrail_extension
    elif StrictVersion(vxrail_version) < StrictVersion(VXRAILNO_V4):
        print_info("VxRail plugin not available for VxRail %s" % vxrail_version)
        print_info("Skip VxRail Plugin installation/registration")
        plugin_check_result = 'Skip'
        return plugin_check_result


""" New VxRail Manager Extension code """


def vxrailIpAddress():
    ip_address = execute_mystic_psql("select ip from settings")
    print "VxRail IP address - %s" % ip_address
    return ip_address


def vxRailServerThumbprint():
    """Retrieve VxRail Manager thumbprint"""
    f_handle = os.popen(
        "openssl s_client -connect localhost:443 < /dev/null 2>/dev/null | openssl x509 -fingerprint -noout -in /dev/stdin")
    try:
        server_thumbprint = f_handle.readlines()[0].strip('\n').split('=')[1]
    except Exception as e:
        print("Error getting thumbprint of localhost")
    print "VxRail Manger self-signed thumbprint is %s" % server_thumbprint
    return server_thumbprint


def register_vxrail_plugin(extension_manager, vxrail_extension_params):
    extension = extension_manager.FindExtension(vxrail_extension_params['extension_key'])

    if extension is not None:
        print_info("VxRail Extension for given key : " + vxrail_extension_params['extension_key'] + " already exists.")
        return

    extension = vim.Extension()
    extension.key = vxrail_extension_params['extension_key']
    extension.company = vxrail_extension_params['company_name']
    extension.version = vxrail_extension_params['version']
    extension.shownInSolutionManager = 1
    extension.lastHeartbeatTime = datetime.utcnow()
    extension.description = vim.Description()
    extension.description.label = vxrail_extension_params['extension_label']
    extension.description.summary = vxrail_extension_params['extension_summary']

    """ set client information """

    client = vim.Extension.ClientInfo()
    client.version = vxrail_extension_params['client_version']
    client.description = vim.Description()
    client.description.label = vxrail_extension_params['client_label']
    client.description.summary = vxrail_extension_params['client_summary']
    client.company = vxrail_extension_params['company_name']
    client.type = "vsphere-client-serenity"
    client.url = vxrail_extension_params['extension_url']

    extension.client.append(client)

    """ set server info """

    server = vim.Extension.ServerInfo()
    server.description = vim.Description()
    server.description.label = vxrail_extension_params['server_label']
    server.description.summary = vxrail_extension_params['server_summary']
    server.company = vxrail_extension_params['company_name']
    server.type = vxrail_extension_params['server_type']
    server.adminEmail.append(vxrail_extension_params['email'])
    server.serverThumbprint = vxrail_extension_params['server_thumbprint']
    server.url = vxrail_extension_params['extension_url']

    extension.server.append(server)

    resource = vim.Extension.ResourceInfo()
    resource.locale = "en"
    resource.module = "task"

    if StrictVersion(vxrail_extension_params['client_version']) >= VXRAILNO_V5:
        """set extension previlege info"""
        """Please note to restart vsphere-ui service after plugin is registered"""
        resource_privilege = vim.Extension.ResourceInfo()
        resource_privilege.locale = "en"
        resource_privilege.module = "auth"

        privilege_group_label = vim.KeyValue()
        privilege_group_label.key = "privilege.vxrail.label"
        privilege_group_label.value = "VxRail"
        resource_privilege.data.append(privilege_group_label)

        privilege_view_label = vim.KeyValue()
        privilege_view_label.key = "privilege.vxrail.View.label"
        privilege_view_label.value = "View VxRail clusters"
        resource_privilege.data.append(privilege_view_label)

        privilege_manage_label = vim.KeyValue()
        privilege_manage_label.key = "privilege.vxrail.Manage.label"
        privilege_manage_label.value = "Manage VxRail clusters"
        resource_privilege.data.append(privilege_manage_label)

        extension.resourceList.append(resource_privilege)

        """set extension privilege list info"""

        extension_view_privilege = vim.Extension.PrivilegeInfo()
        extension_view_privilege.privGroupName = "vxrail"
        extension_view_privilege.privID = extension_view_privilege.privGroupName + "." + "View"

        extension_manage_privilege = vim.Extension.PrivilegeInfo()
        extension_manage_privilege.privGroupName = "vxrail"
        extension_manage_privilege.privID = extension_manage_privilege.privGroupName + "." + "Manage"

        extension.privilegeList.append(extension_view_privilege)
        extension.privilegeList.append(extension_manage_privilege)

        """set extension task info"""
        vc_convert = vim.KeyValue()
        vc_convert.key = vxrail_extension_params['extension_key'] + ".task.vcconversion.label"
        vc_convert.value = "Run VC Conversion"
        resource.data.append(vc_convert)

        vc_convert_task = vim.Extension.TaskTypeInfo()
        vc_convert_task.taskID = vc_convert.key

        extension.taskList.append(vc_convert_task)

    """ set extension resource info """

    cluster_shutdown = vim.KeyValue()
    cluster_shutdown.key = vxrail_extension_params['extension_key'] + ".task.clustershutdown.label"
    cluster_shutdown.value = "Run Cluster Shutdown"
    resource.data.append(cluster_shutdown)

    node_expansion = vim.KeyValue()
    node_expansion.key = vxrail_extension_params['extension_key'] + ".task.nodeexpansion.label"
    node_expansion.value = "Run Node Expansion"
    resource.data.append(node_expansion)

    node_removal = vim.KeyValue()
    node_removal.key = vxrail_extension_params['extension_key'] + ".task.noderemoval.label"
    node_removal.value = "Run Node Removal"
    resource.data.append(node_removal)

    extension.resourceList.append(resource)

    """ set extension task list """

    cluster_shutdown_task = vim.Extension.TaskTypeInfo()
    cluster_shutdown_task.taskID = cluster_shutdown.key
    extension.taskList.append(cluster_shutdown_task)

    node_expansion_task = vim.Extension.TaskTypeInfo()
    node_expansion_task.taskID = node_expansion.key
    extension.taskList.append(node_expansion_task)

    node_removal_task = vim.Extension.TaskTypeInfo()
    node_removal_task.taskID = node_removal.key
    extension.taskList.append(node_removal_task)

    """ register the extension to VC """

    extension_manager.RegisterExtension(extension)
    print_info(
        "Successfully registered VxRail plugin in Targeet vCenter Server as extension %s" % vxrail_extension_params[
            'extension_key'])


def check_vxrail_cusfield(cluster):
    # Iterate the VxRail related custom fields in teh cluster
    print_debug("Checking existing VxRail fields in selected cluster %s" % cluster)
    fields = {}
    custom_field_keys = []
    key_name = []

    for item in cluster.availableField:
        if item.name == vxm_thumbprint_key_name:
            print_debug("Found existing VxRail plugin property %s in cluster %s" % (item.name, cluster))
            custom_field_keys.append(item.key)
            key_name.append(item.name)
            fields[item.name] = item.key
        elif item.name == vxm_ip_key_name:
            print_debug("Found existing VxRail plugin property %s in cluster %s" % (item.name, cluster))
            custom_field_keys.append(item.key)
            key_name.append(item.name)
            fields[item.name] = item.key
        elif item.name == vxm_version_key_name:
            print_debug("Found existing VxRail plugin property %s in cluster %s" % (item.name, cluster))
            custom_field_keys.append(item.key)
            key_name.append(item.name)
            fields[item.name] = item.key

    print_debug("These are exiting custom field keys %s" % custom_field_keys)
    print_debug("These are existing custom field names %s" % key_name)
    print_debug("This is the key and values %s" % fields)

    return fields


def add_cluster_properties(customFieldsManager, cluster, vxrail_extension_params):
    existing_vxm_custom_field = check_vxrail_cusfield(cluster)

    key_value_thumbprint = vim.KeyValue()
    key_value_thumbprint.key = vxm_thumbprint_key_name
    key_value_thumbprint.value = vxrail_extension_params['server_thumbprint']

    key_value_vxm_ip = vim.KeyValue()
    key_value_vxm_ip.key = vxm_ip_key_name
    key_value_vxm_ip.value = vxrail_extension_params['server_ip']

    key_value_vxm_version = vim.KeyValue()
    key_value_vxm_version.key = vxm_version_key_name
    key_value_vxm_version.value = vxrail_extension_params['version']

    key_value_custom_fields = [key_value_thumbprint, key_value_vxm_ip, key_value_vxm_version]

    if len(existing_vxm_custom_field) == 3:
        for item in existing_vxm_custom_field:
            print_debug("Updating custom fields of existing VxRail cluster: %s properties" % cluster)
            print_debug("Updating custom field key: %s, key name: %s" % (item, existing_vxm_custom_field[item]))

            if item == vxm_thumbprint_key_name:
                customFieldsManager.SetField(cluster, existing_vxm_custom_field[item], key_value_thumbprint.value)
                print_info("Successfully reset custom field:%s" % vxm_thumbprint_key_name)
            elif item == vxm_ip_key_name:
                customFieldsManager.SetField(cluster, existing_vxm_custom_field[item], key_value_vxm_ip.value)
                print_info("Successfully reset custom field:%s" % vxm_ip_key_name)
            elif item == vxm_version_key_name:
                customFieldsManager.SetField(cluster, existing_vxm_custom_field[item], key_value_vxm_version.value)
                print_info("Successfully reset custom field:%s" % vxm_version_key_name)

    elif 0 < len(existing_vxm_custom_field) < 3:
        print_debug("Removing existing VxRail related fields")
        for item in existing_vxm_custom_field:
            try:
                customFieldsManager.RemoveCustomFieldDef(existing_vxm_custom_field[item])
            except vim.fault as err:
                print_info("Failed to remove existing fields due to {}".format(err))

    elif len(existing_vxm_custom_field) == 0:
        print_debug("Add new VxRail related custom fields")
        print len(existing_vxm_custom_field)

        for key_value in key_value_custom_fields:
            try:
                keyDef = customFieldsManager.AddCustomFieldDef(key_value.key, vim.ClusterComputeResource)
                customFieldsManager.SetField(cluster, keyDef.key, key_value.value)
                print_info("Successfully added new custom fields:%s" % key_value.key)
            except vim.fault.DuplicateName as err:
                print_info("Failed to add custom field due to {}".format(err))


def unregister(extension_manager, extension_key):
    try:
        extension_manager.UnregisterExtension(extension_key)
        """
        print_error("Please turn to vCenter Server to restart all the services")
        print_info("Using 'service-control --stop --all' inside vCenter Server")
        """

    except vim.fault as err:
        print_info("Failed to un-register VxRail extensions due to {}".format(err))


def update_runtime_properties(file, ext_key, ext_version):
    with open(file, 'r') as fp:
        properties = fp.readlines()

    # Backup the file
    suffix = '.bak_' + datetime.utcnow().strftime("%Y%m%d%H%M%S")
    with open(file + suffix, 'w+') as fp:
        fp.writelines(properties)

    has_ext_key = False
    has_ext_version = False
    for index, item in enumerate(properties):
        if PROPERTY_EXTENSION_KEY in item:
            properties[index] = PROPERTY_EXTENSION_KEY + '=' + ext_key + '\n'
            has_ext_key = True
        elif PROPERTY_EXTENSION_VERSION in item:
            properties[index] = PROPERTY_EXTENSION_VERSION + '=' + ext_version + '\n'
            has_ext_version = True

    if not has_ext_key:
        properties.append(PROPERTY_EXTENSION_KEY + '=' + ext_key + '\n')

    if not has_ext_version:
        properties.append(PROPERTY_EXTENSION_VERSION + '=' + ext_version + '\n')

    with open(file, 'w+') as fp:
        fp.writelines(properties)


def menu_input(vc_ip):
    """get user input """
    mvcip = raw_input('Please input vCenter IP address[default:' + vc_ip + ']:') or vc_ip
    muser = raw_input(
        'Please input vCenter Server Admin username [default:administrator@vsphere.local]:') or "administrator@vsphere.local"
    mpwd = getpass.getpass('Please input vCenter Admin password:')

    args = {}
    args['vcip'] = mvcip
    args['username'] = muser
    args['password'] = mpwd

    return args


def main():
    """
    Simple command-line program to register VxRail Manager as extension to VC.
    """
    vc_ip = execute_mystic_psql("select vcenter_ip from settings")
    args = menu_input(vc_ip)

    vxrail_version = vxrailversion()
    if StrictVersion(vxrail_version) <= StrictVersion(VXRAILNO_V4):
        vxrail_extension_version = VXRAILNO_V4
    elif StrictVersion(vxrail_version) > StrictVersion(VXRAILNO_V4):
        vxrail_extension_version = vxrail_version


    vxrail_ip = vxrailIpAddress()

    # Connect to vCenter Server
    ssl._create_default_https_context = ssl._create_unverified_context
    print_info("Connecting to VC: %s" % (args['vcip']))
    si = Connect(host=args['vcip'], user=args['username'],
                 pwd=args['password'], version=VIM_VERSION)
    atexit.register(Disconnect, si)

    content = si.RetrieveContent()

    collect_info = {}
    collect_info['VC_IP'] = args["vcip"]
    collect_info['VC_USER'] = args["username"]
    collect_info['VC_PWD'] = args["password"]

    # Parameters prepared for VxRail Manager extensions:
    vxrail_plugin_status = {}
    vxrail_extension_params = {}
    vxrail_extension_params['version'] = vxrail_extension_version
    vxrail_extension_params['company_name'] = "VMware Inc."
    vxrail_extension_params['email'] = "noreply@vmware.com"
    vxrail_extension_params['server_thumbprint'] = vxRailServerThumbprint()
    vxrail_extension_params['server_ip'] = vxrail_ip
    vxrail_extension_params['client_version'] = vxrail_extension_version

    if StrictVersion(vxrail_version) <= StrictVersion(VXRAILNO_V4):
        vxrail_extension_params['client_label'] = "VxRail " + vxrail_extension_version + " client plugin for vCenter."
        vxrail_extension_params['client_summary'] = "vSphere Client Extension for VxRail clusters on v" + vxrail_extension_version
        vxrail_extension_params['extension_label'] = "VxRail Extension for " + vxrail_extension_version
        vxrail_extension_params['extension_summary'] = "vSphere Client Extension for VxRail clusters on v" + vxrail_extension_version
    elif StrictVersion(vxrail_version) > StrictVersion(VXRAILNO_V4):
        vxrail_extension_params['extension_label'] = "VxRail HTML5 Client Plugin"
        vxrail_extension_params['extension_summary'] = "vSphere Client Extension for VxRail Clusters"
        vxrail_extension_params['client_label'] = "VxRail HTML5 Client Plugin"
        vxrail_extension_params['client_summary'] = "vSphere Client Extension for VxRail Clusters"

    vxrail_extension_params['server_type'] = "HTTPS"
    vxrail_extension_params['server_label'] = "VxRail Manager server"
    vxrail_extension_params['server_summary'] = "VxRail Manager server"

    vxrail_extension_params['extension_url'] = "https://" + vxrail_extension_params["server_ip"] + "/download/vxrail-plugin"
    vxrail_extension_params['extension_key'] = "com.vmware.vxrail"

    ent_datacenter = collect_datacenter(content, collect_info)
    vxrail_cluster = collect_cluster(ent_datacenter, collect_info)

    vxrail_extension_params['cluster_moId'] = collect_info['cluster_id']

    # Uninstall and Register VxRail Manager Plugin to Target vCenter Server
    while True:
        plugin_check, vxrail_extensions = vxrail_plugin_check(vxrail_version, si.content.extensionManager)

        if plugin_check == 'Pass':
            try:
                register_vxrail_plugin(si.content.extensionManager, vxrail_extension_params)
                print_info(
                    "Successfully registered VxRail Plugin as extension %s" % vxrail_extension_params['extension_key'])

                add_cluster_properties(si.content.customFieldsManager, vxrail_cluster, vxrail_extension_params)
                print_info("Successfully added new custom fields to VxRail cluster %s" % vxrail_cluster)
                print_info("Successfully registered VxRail plugin!")

                # Update runtime.properties
                update_runtime_properties(RUNTIME_PROPERTIES_FILE, vxrail_extension_params['extension_key'],
                                          vxrail_extension_params['version'])
                print_info("Updating runtime.properties Succeeded!")
                sys.exit(0)

            except vmodl.fault.InvalidArgument as err:
                print_error("Failed to register extension. {}".format(err))
                sys.exit(1)
        elif plugin_check == 'Installed':
            # As it's for internal VxRail testing, any existing VxRail plugin will be removed. So the following is not necessary.
            # print_debug("Configuring selected VxRail Cluster %s to work with VxRail plugin" % vxrail_cluster)
            # try:
            #    add_cluster_properties(si.content.customFieldsManager, vxrail_cluster, vxrail_extension_params)
            # except vmodl.fault.InvalidArgument as err:
            #    print_error("Failed to configure customFields. {}".format(err))
            #    return -1
            act = raw_input('Are you sure that you want to un-register ALL VxRail extensions? Y or N:')

            if act == 'Y' or act == 'y':
                print_error("Un-registering VxRail extensions")
                for item in vxrail_extensions:
                    unregister(si.content.extensionManager, item)
                    print_debug("Un-register VxRail plugin %s success" % vxrail_extensions)

            else:
                print_info("Exit")
                sys.exit(0)

        elif plugin_check == 'Skip':
            print_info("This VxRail version is not compatible with VxRail plugin. Exiting...")
            sys.exit(1)

        print_info("Restarting service vmware-marvin, please wait...")
        os.popen("service vmware-marvin restart")
        print_info("Restarting service runjars, please wait...")
        os.popen("service runjars restart")
        print_info("Restarting service vmware-loudmouth, please wait...")
        os.popen("service vmware-loudmouth restart")

# Start program
if __name__ == "__main__":
    main()